from eigenwerte_1 import *

""" Modifiziertes Signum, das 1 fuer Zahlen >= 0 gibt, -1 sonst
    Arguments:
    xx -- zu testendes Ding
"""
def my_sign(xx):
    return 1. if abs(xx) == 0 else xx/abs(xx)

""" Testest, ob Matrix auch Hessenbergform hat
    Arguments:
    AA -- Matrix
    eps -- Genauigkeit
"""
def is_hessenberg(AA, eps=1e-14):
    return np.max(np.abs(np.tril(AA, k=-2))) < eps

""" Testet, ob Matrix unitaer ist
    Arguments:
    AA -- Matrix
    eps -- Genauigkeit
"""
def is_unitary(AA, eps=1e-14):
    return np.max(np.abs(np.eye(np.shape(AA)[0])-AA.dot(np.conjugate(AA.T)))) < eps and np.max(np.abs(np.eye(np.shape(AA)[0])-np.conjugate(AA.T).dot(AA))) < eps

""" Testet ob QQ.T.dot(AA.dot(QQ)) == BB
    Arguments:
    AA -- Matrix
    BB -- Matrix
    QQ -- Matrix
    eps -- Genauigkeit
"""
def are_similar(AA, BB, QQ, eps=1e-14):
    return np.max(np.abs(np.conjugate(QQ.T).dot(AA.dot(QQ))-BB)) < eps

""" Testet, ob Matrix tridiagonal ist
    Arguments:
    AA -- Matrix
    eps -- Genauigkeit
"""
def is_tridiagonal(AA, eps=1e-14):
    return np.max(np.abs(np.tril(AA, k=-2))) < eps and np.max(np.abs(np.triu(AA, k=2))) < eps

""" Transformiert eine Matrix AA mit Householder-Spiegelungen auf obere Hessenberg-Form
    Arguments:
    AA -- zu transformierende Matrix
    return_QQ -- ob Transformationsmatrix
    Rueckgabe
    HH -- transformierte Matrix in Hessenbergform
    QQ -- Transformationsmatrix sodass: QQ.T.dot(AA.dot(QQ)) = HH
"""
def to_hessenberg(AA, return_QQ=False):
    if sp.issparse(AA):
        HH = AA.toarray()
    else:
        HH = AA.copy()
    QQ = np.eye(np.shape(AA)[0])
    if np.iscomplexobj(AA):
        QQ = QQ+0j
    for ii in                             :
        # Spiegelungsvektor konstruieren
        vv =                      
                                                   
                                 
        # Transformationsmatrix update, auf effizienten Anwendung von np.outer(vv, vv) achten!
        QQ                                                
        # Spiegelung von links anwenden
        HH                                       
                                                                  
        # Spiegelung von rechts anwenden
        HH                                                
    # Rueckgabe
    if return_QQ:
        return HH, QQ
    else:
        return HH

""" Gibt 2x2 Givens Matrix, die JJ = AA[jj,kk], mit Hilfe von II = AA[ii, kk] eliminiert,
    also die (verkleinerte) Givens-Matrix G(ii, jj)
    Arguments:
    II -- Wert mit dem eliminiert werden soll
    JJ -- Wert der eliminiert werden soll
"""
def get_givens(II, JJ):
    # Cosinus und Sinus berechnen
                                
                                                                                             
         
                                                                                             
    # Rueckgabe als 2x2 Matrix
    return np.array([[cc, ss], [-ss, cc]])

""" Modifizierte QR-Zerlegung fuer Matrizen in Hessenbergform mit Givens-Rotationen
    Arguments:
    HH -- Matrix in Hessenbergform, die in QQ.dot(RR) == HH zerlegt wird
    only_SS -- nur RR.dot(QQ) zurueckgeben
    Rueckgabe
    QQ -- Transformation
    RR -- rechte Seite
    SS -- RR.dot(QQ)
"""
def hessenberg_qr(HH, only_SS=False):
    RR = HH.copy()
    # Platz zum Speichern der Rotationen
    if np.iscomplexobj(RR):
        rotations = np.array([np.eye(2)+0j for ii in xrange(np.shape(HH)[0]-1)])
    else:
        rotations = np.array([np.eye(2) for ii in xrange(np.shape(HH)[0]-1)])
    # Rotationen von links anwenden, auf effiziente Anwendung achten, nur 2 Zeilen aendern!
    for ii in xrange(0, np.shape(HH)[0]-1):
        # Rotation berechnen und speichern
                                                         
        # Rotation von links anwenden
                                                            
    SS = RR.copy()
    # Platz fuer QQ
                           
                                       
         
                                    
    # Rotationen von rechts anwenden, auf effizienten Anwendung achten, nur 2 Spalten aendern
    for ii in xrange(0, np.shape(HH)[0]-1):
        SS                                                                      
        QQ                                                                      
    # Rueckgabe
    if only_SS:
        return SS
    else:
        return QQ, RR, SS

""" QR-Verfahren mit Shift
    Arguments:
    AA -- Matrix
    eps -- gewuenschte Genauigkeit, maximaler Betrag von Subdiagonalelementen
    Rueckgabe:
    VV -- Eigenwerte
    WW -- Eigenvektoren
"""
def qr_shift(AA, eps=1e-7):
    # Hessenbergtransformation
    HH =                  
    # Schleife
    err = np.max(np.abs(np.diag(HH,-1)))
    while err > eps:
        # Shift 
        SS =                                  
        # QR Zerlegung
        QQ, RR, HH = hessenberg_qr(HH+SS)
        # Shift Teil 2
        HH      
                                            
    return np.diag(RR), QQ

""" Malt geforderte Bilder mit Einschliessungen und Eigenwerten
"""
def plot_eigenvalue_stuff(AA, eps=1e-7, fig=None):
    if fig == None:
        fig = plt.figure()
    ax = fig.add_subplot(111)
    # Einschliessungen malen
    plot_gerschgorin(AA, ax)
    plot_bendixson(AA, ax)
    ax.set_xlabel(r'$\mathbb{R}$')
    ax.set_ylabel(r'$\mathrm{i} \mathbb{R}$')
    # Eigenwerte berechnen
                        
    # Eigenwerte malen
                                           
    # zeigen
    fig.show()

""" Testfaelle und Aufgaben
"""
if __name__ == '__main__':
    eps = 1e-7
    # Test
    AA = np.array([[4., 0., 3.], [0., -1., 1.], [1., 1., 0.]])
    HH, QQ = to_hessenberg(AA, True)
    print "Test 0: ", is_hessenberg(HH), is_unitary(QQ), are_similar(AA, HH, QQ)
    plot_eigenvalue_stuff(AA)
    plt.gcf().savefig('test_0.pdf')

    testnn = 4
    BB = np.diag(rand(testnn))+np.diag(rand(testnn)*1j)+np.diag(rand(testnn-1), 1)+np.diag(rand(testnn-1),-1)
    HH, QQ = to_hessenberg(BB, True)
    print "Test 1: ", is_hessenberg(HH), is_unitary(QQ), are_similar(BB, HH, QQ)
    plot_eigenvalue_stuff(BB)
    plt.gcf().savefig('test_1.pdf')

    CC = BB+np.conjugate(BB.T)
    HH, QQ = to_hessenberg(CC, True)
    print "Test 2: ", is_tridiagonal(HH), is_unitary(QQ), are_similar(CC, HH, QQ)
    plot_eigenvalue_stuff(CC)
    plt.gcf().savefig('test_2.pdf')

    # Aufgabe
    for ii in xrange(10, 21, 10):
        AA = get_gram(ii)
        plot_eigenvalue_stuff(AA)
        plt.gcf().savefig(str(ii)+'_eigenwerte.pdf')
